// Copyright © 2016 JABT
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions: The above copyright
// notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

import UIKit

open class Animation {
    private static let timeResolution = 0.01 as TimeInterval

    ///  Specifies the supported animation curves.
    public enum Curve {
        /// A linear animation curve causes an animation to occur evenly over its duration.
        case linear
        /// An ease-out curve causes the animation to begin quickly, and then slow down as it completes.
        case easeOut
        /// An ease-in curve causes the animation to begin slowly, and then speed up as it progresses.
        case easeIn
        /// An ease-in ease-out curve causes the animation to begin slowly, accelerate through the middle of its duration, and then slow again before completing. This is the default curve for most animations.
        case easeInOut
        /// A stepLeft curve causes the animation to happen instantaneously at the beginning.
        case stepLeft
        /// A stepLeft curve causes the animation to happen instantaneously in the middle.
        case stepMiddle
        /// A stepLeft curve causes the animation to happen instantaneously at the end.
        case stepRight

        var isStep: Bool {
            switch self {
            case .stepRight, .stepMiddle, .stepLeft:
                return true
            case .easeIn, .easeOut, .easeInOut, .linear:
                return false
            }
        }
    }

    /// The amount of time to way before executing the animation.
    open var delay: TimeInterval = 0

    /// The duration of the animation, measured in seconds.
    open var duration: TimeInterval = 1

    /// The animation curve that the receiver will apply to the changes it is supposed to animate.
    open var curve: Curve = .easeInOut

    /// A block with animations to execute.
    open var animations: () -> Void

    static var stack = [Animation]()
    static var currentAnimation: Animation? {
        return stack.last
    }

    public init(delay: TimeInterval, duration: TimeInterval, curve: Curve, animations: @escaping () -> Void) {
        self.curve = curve
        self.animations = animations
        self.duration = curve.isStep ? 0 : duration

        switch curve {
        case .linear, .easeIn, .easeOut, .easeInOut:
            self.delay = delay
        case .stepRight:
            self.delay = delay + duration
        case .stepLeft:
            self.delay = delay + Animation.timeResolution
        case .stepMiddle:
            self.delay = delay + duration/2.0
        }
    }

    open var timingFunction: CAMediaTimingFunction? {
        switch curve {
        case .linear:
            return CAMediaTimingFunction(name: kCAMediaTimingFunctionLinear)
        case .easeOut:
            return CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseOut)
        case .easeIn:
            return CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseIn)
        case .easeInOut:
            return CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut)
        case .stepLeft, .stepMiddle, .stepRight:
            return nil
        }
    }

    open var options: UIViewAnimationOptions {
        switch curve {
        case .linear:
            return .curveLinear
        case .easeOut:
            return .curveEaseOut
        case .easeIn:
            return .curveEaseIn
        case .easeInOut:
            return .curveEaseInOut
        case .stepLeft:
            return []
        case .stepMiddle:
            return []
        case .stepRight:
            return []
        }
    }

    /// Initiates the changes specified in the receivers `animations` block.
    open func animate() {
        let time = DispatchTime.now() + Double(Int64(delay * Double(NSEC_PER_SEC))) / Double(NSEC_PER_SEC)
        DispatchQueue.main.asyncAfter(deadline: time) {
            UIView.animate(withDuration: self.duration, delay: 0, options: self.options, animations: {
                Animation.stack.append(self)
                self.doInTransaction(self.animations)
                Animation.stack.removeLast()
            }, completion:nil)
        }
    }

    fileprivate func doInTransaction(_ action: () -> Void) {
        CATransaction.begin()
        CATransaction.setAnimationDuration(duration)
        if let timingFunction = timingFunction {
            CATransaction.setAnimationTimingFunction(timingFunction)
        }
        action()
        CATransaction.commit()
    }
}

extension CABasicAnimation {
    public func configureOptions() {
        self.fillMode = kCAFillModeBoth
        self.isRemovedOnCompletion = false
    }
}

extension CATransaction {
    static func suppressAnimations(actions: () -> Void) {
        begin()
        setAnimationDuration(0)
        actions()
        commit()
    }
}
